from typing import IO

from wgrd_cons_tools.utils import *


def stream_write_dds(stream: IO[bytes], w: int, h: int, format: str, mipmaps: list[IO[bytes]]):
    # FIXME: put these in enum?
    DXGI_FORMAT_R16_UNORM = 56
    DXGI_FORMAT_R8_TYPELESS = 60
    DXGI_FORMAT_BC1_UNORM = 71
    DXGI_FORMAT_BC1_UNORM_SRGB = 72
    DXGI_FORMAT_BC3_TYPELESS = 76
    DXGI_FORMAT_BC3_UNORM = 77
    DXGI_FORMAT_BC3_UNORM_SRGB = 78
    DXGI_FORMAT_B8G8R8A8_UNORM = 87
    DXGI_FORMAT_B8G8R8A8_UNORM_SRGB = 91

    # these are specific to the game
    ddsFormats = {
        "DXT1": (DXGI_FORMAT_BC1_UNORM, None),
        "DXT1_SRGB": (DXGI_FORMAT_BC1_UNORM_SRGB, None),
        "DXT5": (DXGI_FORMAT_BC3_TYPELESS, None),
        "DXT5_LIN": (DXGI_FORMAT_BC3_UNORM, None),
        "DXT5_SRGB": (DXGI_FORMAT_BC3_UNORM_SRGB, None),
        'L8': (DXGI_FORMAT_R8_TYPELESS, None),
        "L16_LIN": (DXGI_FORMAT_R16_UNORM, None),
        "A8R8G8B8_LIN": (DXGI_FORMAT_B8G8R8A8_UNORM, None),
        'A8R8G8B8_SRGB': (DXGI_FORMAT_B8G8R8A8_UNORM_SRGB, None)
    }
    ddsFormat = ddsFormats[format]

    # Write https://learn.microsoft.com/en-us/windows/win32/direct3ddds/dds-header

    # FIXME: put these in enum?
    DDSD_CAPS = 0x1
    DDSD_HEIGHT = 0x2
    DDSD_WIDTH = 0x4
    DDSD_PIXELFORMAT = 0x1000
    DDSD_MIPMAPCOUNT = 0x20000
    DDSCAPS_MIPMAP = 0x400000
    DDSCAPS_TEXTURE = 0x1000
    DDSCAPS_COMPLEX = 0x8

    stream.write(b'DDS ')
    write32(stream, 124) # DWORD dwSize;
    write32(stream, DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT | DDSD_MIPMAPCOUNT) # DWORD dwFlags;
    write32(stream, h) # DWORD dwHeight;
    write32(stream, w) # DWORD dwWidth;
    write32(stream, 0) # DWORD dwPitchOrLinearSize;
    write32(stream, 0) # DWORD dwDepth;
    write32(stream, len(mipmaps)) # DWORD dwMipMapCount;
    for _ in range(11):
        write32(stream, 0) # DWORD dwReserved1[11];

    if True:
        # Write DDS_PIXELFORMAT ddspf;
        DDPF_FOURCC = 0x4
        write32(stream, 32) # DWORD dwSize;
        write32(stream, DDPF_FOURCC) # DWORD dwFlags;
        stream.write(b'DX10') # DWORD dwFourCC;
        write32(stream, 0) # DWORD dwRGBBitCount;
        write32(stream, 0) # DWORD dwRBitMask;
        write32(stream, 0) # DWORD dwGBitMask;
        write32(stream, 0) # DWORD dwBBitMask;
        write32(stream, 0) # DWORD dwABitMask;

    write32(stream, DDSCAPS_TEXTURE | DDSCAPS_MIPMAP | DDSCAPS_COMPLEX) # DWORD dwCaps;
    write32(stream, 0) # DWORD dwCaps2;
    write32(stream, 0) # DWORD dwCaps3;
    write32(stream, 0) # DWORD dwCaps4;
    write32(stream, 0) # DWORD dwReserved2;

    # Write https://learn.microsoft.com/en-us/windows/win32/direct3ddds/dds-header-dxt10

    # FIXME: put these in enum?
    D3D10_RESOURCE_DIMENSION_TEXTURE2D = 3
    DDS_ALPHA_MODE_STRAIGHT = 1

    write32(stream, ddsFormat[0]) # DXGI_FORMAT dxgiFormat;
    write32(stream, D3D10_RESOURCE_DIMENSION_TEXTURE2D) # D3D10_RESOURCE_DIMENSION resourceDimension;
    write32(stream, 0) # UINT miscFlag;
    write32(stream, 0) # UINT arraySize; #FIXME: Might be bad?
    write32(stream, DDS_ALPHA_MODE_STRAIGHT) # UINT miscFlags2;

    # DDS mipmap order is different from TGV
    for mipmap_file in mipmaps:
        stream.write(mipmap_file.getbuffer())




#!/usr/bin/env python3

import os
import sys
import argparse
import pathlib

from wgrd_cons_parsers.edat import EDat


def create_vfs(dat_file_paths: list[str], vfs_dict: dict[str, pathlib.Path] = None) -> dict[str, pathlib.Path]:
    """
    Create a virtual file system from a list of dat files.
    :param dat_file_paths: list of dat files to be searched
    :param vfs_dict: vfs_dict to be appended to, so it is possible to chain multiple create_vfs calls
    :return: vfs_dict containing all the vfs filepaths from the dat files
    """
    if vfs_dict is None:
        vfs_dict = {}

    idx = 0
    for filepath in dat_file_paths:
        print(f"parsing {filepath}")
        data = EDat.parse_file(filepath, _cons_xml_filesdictionary_parse_files=False)
        for vfspath, meta in data.files.items():
            vfs_dict[vfspath] = (os.path.abspath(filepath), meta.offset + data.offset_data, meta.size, meta.checksum, idx)
            idx += 1

    return vfs_dict


def get_dat_paths(wgrd_path: pathlib.Path) -> list[str]:
    """
    Recursively search for .dat files in the wgrd_path directory.
    :param wgrd_path: Path to the directory containing all the dat files
    :return: list of dat files, sorted by name
    """
    dat_file_paths = []

    for root, dirs, files in os.walk(wgrd_path):
        for file in files:
            if file.endswith(".dat"):
                filepath = os.path.join(root, file)
                dat_file_paths.append(filepath)

    return sorted(dat_file_paths)


def create_vfs_tree(vfs_dict: dict[str, pathlib.Path]) -> dict:
    """
    Create a vfs tree from a vfs_dict.
    :param vfs_dict: vfs_dict mapping vfs paths to dat filepaths
    :return: nested directory structure of the vfs mapping to dat filepaths in the end
    """
    vfs_tree = {}

    for vfs_path, fs_path in vfs_dict.items():
        parts = vfs_path.split("\\")
        last = vfs_tree
        for part in parts[:-1]:
            last.setdefault(part, {})
            last = last[part]
        last[parts[-1]] = fs_path
    
    return vfs_tree


def search_vfs_tree(vfs_dict: dict[str, pathlib.Path], search_string: str) -> dict:
    """
    Search a vfs dict (not the tree) for a search string.
    :param vfs_dict:
    :param search_string:
    :return: a vfs_tree only containing the paths that contain the search string
    """
    search_string.strip()
    search_string = search_string.replace("\\", "/")
    return {k: v for k, v in vfs_dict.items() if search_string in k.replace("\\", "/")}


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("wgrd_path", type=pathlib.Path, help="path to the wgrd game files")
    #parser.add_argument("output_path", type=pathlib.Path, help="unpack directory", nargs='?', default="./out/")

    args = parser.parse_args()

    print(create_vfs_tree(create_vfs(get_dat_paths(args.wgrd_path))))



import pathlib
import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("wgrd_exe", type=pathlib.Path,
                        help="path to the vanilla Wargame3.exe")
    parser.add_argument("-r", "--revision", type=int, default=80001)
    parser.add_argument("-o", "--output", default="out/Wargame3_modded.exe", type=pathlib.Path,
                        help="path to the output Wargame3.exe")
    args = parser.parse_args()

    position = 0x01520213

    with open(args.output, "wb") as patched:
        with open(args.wgrd_exe, "rb") as original:
            patched.write(original.read(position))
            original_revision = original.read(10)
            new_revision = str(args.revision)
            assert(len(new_revision) <= 10)
            new_revision = (10-len(new_revision)) * "0" + new_revision

            patched.write(new_revision.encode("ascii"))
            patched.write(original.read())



from wgrd_cons_tools.create_vfs import *

def test_search_vfs_tree():
    test_tree = {"a\\b": "test1.dat", "a\\c": "test2.dat", "a\\b\\c": "test3.dat"}
    assert(search_vfs_tree(test_tree, "a/b") == {"a\\b": "test1.dat", "a\\b\\c": "test3.dat"})
    assert(search_vfs_tree(test_tree, "a\\c") == {"a\\c": "test2.dat"})




#!/usr/bin/env python3
import sys

from wgrd_cons_parsers.ess import *
from wgrd_cons_parsers.sformat import *

import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("path", type=pathlib.Path, help="path to the ess file")
    parser.add_argument("-o", "--output", type=str, default="",
                        help="path to the output sformat file")
    args = parser.parse_args()

    if not len(args.output):
        args.output = os.path.basename(args.path) + ".sformat"

    essfile = open(args.path, "rb")
    essdata = essfile.read()

    ess_header = Ess.parse(essdata)

    sformat = Container(
              unk0=0x6,
              isShort="true",
              channelCount=ess_header.channels,
              frameSize=ess_header.channels*0x200,
              sampleRate=ess_header.samplerate,
              frameCount=ess_header.frameCount,
              unk4=2,
              essLength=len(essdata),
              loopStart=ess_header.loopStart,
              loopEnd=ess_header.loopEnd,
              data=None)

    sformatdata = SFormat.build(sformat)

    sformatfile = open(args.output, "wb")
    sformatfile.write(sformatdata)



from wgrd_cons_parsers.edat import EDat

import os, pathlib
import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-w", "--wgrd_path", type=pathlib.Path,
                        help="path to the vanilla game files, e.g. C:/Program Files (x86)/Steam/steamapps/common/Wargame Red Dragon/Data/WARGAME/PC")
    parser.add_argument("-m", "--modded_path", type=pathlib.Path,
                        help="path to the modded game files")
    parser.add_argument("-o", "--output", default="out/", type=pathlib.Path,
                        help="path to the output directory")
    args = parser.parse_args()

    output_edat_files = {}

    for root, dirs, files in os.walk(args.modded_path):
        for file in files:
            if file.endswith(".dat"):
                rel_root = os.path.relpath(root, args.modded_path)
                rel_filepath = os.path.join(rel_root, file)
                original_filepath = os.path.join(args.wgrd_path, rel_filepath)
                modded_filepath = os.path.join(args.modded_path, rel_filepath)

                print("parsing file: ", modded_filepath)

                extra = {"_cons_xml_filesdictionary_alignment": False,
                         "_cons_xml_filesdictionary_disable_checks": True}
                original_edat = EDat.parse_file(original_filepath, **extra)
                modded_edat = EDat.parse_file(modded_filepath, **extra)

                output_edat_file = output_edat_files.get(file, {"sectorSize": original_edat["sectorSize"], "files": {}})
                for k, v in modded_edat["files"].items():
                    if k in original_edat["files"]:
                        if v != original_edat["files"][k]:
                            print(k)
                            output_edat_file["files"][k] = v
                        #else:
                        #    output_edat_file["files"][k] = v
                output_edat_files[file] = output_edat_file

    os.makedirs(args.output, exist_ok=True)
    for k, v in output_edat_files.items():
        preprocessed, _ = EDat.preprocess(v)
        output_edat = EDat.build_file(preprocessed, os.path.join(args.output, k))

